from __future__ import annotations

import numbers

import polars as pl
from hypothesis import given, settings
from polars.testing.parametric import dataframes, column
import pytest
from river import compose, datasets, preprocessing, stream
from river.datasets.synth import AnomalySine

from river_rrcf.rrcf import RobustRandomCutForest


def model1():
    rrcf = RobustRandomCutForest(num_trees=4, tree_size=32, random_state=42)
    scaler = compose.SelectType(numbers.Number) | preprocessing.StandardScaler()  # type: ignore
    ordinal = compose.SelectType(str) | preprocessing.OrdinalEncoder()  # type: ignore
    return compose.Pipeline(
        scaler + ordinal,
        rrcf,
    )


def model2():
    rrcf = RobustRandomCutForest(num_trees=4, tree_size=32, random_state=42)
    scaler = compose.Select("integer", "float") | preprocessing.StandardScaler()  # type: ignore
    ordinal = compose.Select("string") | preprocessing.OrdinalEncoder()  # type: ignore
    return compose.Pipeline(
        scaler + ordinal,
        rrcf,
    )


def test_synth():
    rrcf = model1()
    dataset = AnomalySine(n_samples=100, n_anomalies=10, seed=42)

    for x, _ in dataset:
        score = rrcf.score_one(x)
        rrcf.learn_one(x)
        assert isinstance(score, float)


def test_credit_card():
    rrcf = model1()
    dataset = datasets.CreditCard().take(100)

    for x, _ in dataset:
        score = rrcf.score_one(x)
        rrcf.learn_one(x)
        assert isinstance(score, float)


@given(
    data=dataframes(
        allowed_dtypes=[pl.Int64, pl.Float32, pl.String],
        min_cols=1,
        max_cols=5,
        min_size=1,
        max_size=10,
        allow_null=False,
    )
)
@settings(deadline=None)
def test_dataframes_model1(data: pl.DataFrame):
    rrcf = model1()

    dataset = stream.iter_polars(data)
    for x, _ in dataset:
        score = rrcf.score_one(x)
        rrcf.learn_one(x)
        assert isinstance(score, float)


def test_unseen_feature_model1():
    rrcf = model1()
    rrcf.learn_one({"a": None, "b": None})

    assert rrcf["RobustRandomCutForest"]._keys is None

    rrcf.learn_one({"a": 0.0, "b": None})
    assert rrcf["RobustRandomCutForest"]._keys == ["a"]

    with pytest.raises(ValueError, match="Unseen features:"):
        rrcf.score_one({"a": 0.0, "b": "asdf"})


@given(
    data=dataframes(
        [
            column("integer", dtype=pl.Int64),
            column("float", dtype=pl.Float32),
            column("string", dtype=pl.String),
        ],
        min_cols=1,
        max_cols=5,
        min_size=1,
        max_size=10,
        allow_null=True,
    )
)
@settings(deadline=None)
def test_dataframes_model2(data: pl.DataFrame):
    rrcf = model2()

    dataset = stream.iter_polars(data)
    for x, _ in dataset:
        score = rrcf.score_one(x)
        rrcf.learn_one(x)
        assert isinstance(score, float)



from __future__ import annotations

from collections import deque

import numpy as np
from river.anomaly.base import AnomalyDetector

from river_rrcf._vendor import rrcf

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any, SupportsFloat

    from numpy.typing import NDArray


FLOAT32_MAX = 3.4028234663852886e38
FLOAT32_MIN = -3.4028234663852886e38


class RobustRandomCutForest(AnomalyDetector):
    """
    A Robust Random Cut Forest (RRCF) for anomaly detection.

    Parameters
    ----------
    num_trees : int, default=40
        The number of trees in the forest.
    tree_size : int, default=256
        The maximum size of each tree.
    shingle_size : int, default=1
        The size of the shingle (window) for time series data.
    random_state : int | None, default=None
        Random seed for each tree in the forest. (>= 0 if int)

    Attributes
    ----------
    num_trees : int
        The number of trees in the forest.
    tree_size : int
        The maximum size of each tree.
    shingle_size : int
        The size of the shingle (window) for time series data.
    forest : list
        A list of RCTree instances representing the forest.
    _index : int
        The current index of the data point being processed.
    _keys : list or None
        The keys of the input data mapping.
    _shingle : deque or None
        A deque representing the shingle for time series data.

    Methods
    -------
    _preprocess(x)
        Preprocesses the input data mapping into a numpy array.
    _init_shingle(arr)
        Initializes the shingle with zeros.
    learn_one(x)
        Learns from a single data point.
    score_one(x)
        Scores a single data point for anomaly detection.
    """

    def __init__(
        self,
        num_trees: int = 40,
        tree_size: int = 256,
        *,
        shingle_size: int = 1,
        random_state: int | None = None,
    ):
        self.num_trees = num_trees
        self.tree_size = tree_size
        self.shingle_size = shingle_size
        self.random_state = random_state

        if random_state is not None:
            self.forest = [
                rrcf.RCTree(random_state=random_state + i) for i in range(num_trees)
            ]
        else:
            self.forest = [rrcf.RCTree(random_state=None) for _ in range(num_trees)]

        self._index = 0
        self._keys: list[Any] | None = None

        self._shingle: deque[NDArray[np.float64]] | None = None

    def _preprocess(self, x: Mapping[Any, SupportsFloat]) -> NDArray[np.float64]:
        """
        Preprocesses the input mapping by converting its values to a numpy array of floats.

        Parameters
        ----------
        x : Mapping[Any, SupportsFloat]
            The input mapping where keys are any hashable type and values are convertible to float.

        Returns
        -------
        NDArray[np.float64]
            An 1-d numpy array of floats corresponding to the values of the input mapping.

        Raises
        ------
        ValueError
            If there are keys in the input mapping that were not expected based on the initial keys.
        """
        if self._keys is None:
            self._keys = list(x)

        xx = dict(x)
        items = [float(xx.pop(k, 0.0)) for k in self._keys]

        if xx:
            remain = ", ".join(str(r) for r in xx)
            expected = ", ".join(str(k) for k in self._keys)
            msg = f"Unseen features: {remain}\n       expected: {expected}"
            raise ValueError(msg)

        arr = np.array(items, dtype=np.float64)
        return np.nan_to_num(arr, copy=False).clip(FLOAT32_MIN, FLOAT32_MAX)

    def _init_shingle(self, arr: NDArray[np.float64]) -> deque[NDArray[np.float64]]:
        a = np.zeros_like(arr)
        return deque([a] * self.shingle_size, maxlen=self.shingle_size)

    def learn_one(self, x: Mapping[Any, SupportsFloat]) -> None:
        """
        Learn a single data point.

        Note: The input values are restricted to the float32 range.

        Parameters
        ----------
        x : Mapping[Any, SupportsFloat]
            The data point to learn.

        """
        if len(x) == 0 and not self._keys:
            return

        xx = self._preprocess(x)

        if not self._shingle:
            self._shingle = self._init_shingle(xx)

        self._shingle.append(xx)

        for tree in self.forest:
            if len(tree.leaves) >= self.tree_size:
                tree.forget_point(self._index - self.tree_size)
            arr = np.concatenate(self._shingle)
            tree.insert_point(arr, index=self._index)

        self._index += 1

    def score_one(self, x: Mapping[Any, SupportsFloat]) -> float:
        """
        Computes the anomaly score for a single data point.

        Parameters
        ----------
        x : Mapping[Any, SupportsFloat]
            The data point to be scored. It should be a mapping where keys are hashable and values are floats.

        Returns
        -------
        float
            The computed anomaly score for the given data point.

        Notes
        -----
        - If the input data point is empty and the forest is not initialized, this method returns 0.0.
        - The anomaly score is calculated as the average co-displacement (codisp) across all trees in the forest.
        """
        if len(x) == 0 and not self._keys:
            return 0.0

        xx = self._preprocess(x)
        i = -1

        if not self._shingle:
            self._shingle = self._init_shingle(xx)

        data = [*self._shingle, xx]
        if len(data) > self.shingle_size:
            data = data[-self.shingle_size :]

        for tree in self.forest:
            tree.insert_point(data, index=i)

        avg_codisp = np.mean([tree.codisp(i) for tree in self.forest], dtype=np.float64)

        for tree in self.forest:
            tree.forget_point(i)

        return avg_codisp.item()



import numpy as np
import pytest
from hypothesis import given, assume, settings
from hypothesis import strategies as st

from river_rrcf.rrcf import RobustRandomCutForest


@given(
    st.dictionaries(
        st.text(),
        st.one_of(st.floats(width=32), st.integers()),
        min_size=1,
    )
)
def test_preprocess_valid_input(data: dict[str, int | float]):
    rrcf = RobustRandomCutForest()

    expected_output = np.array(list(data.values()), dtype=np.float64)
    expected_output = np.nan_to_num(expected_output)
    float32_max = np.finfo(np.float32).max
    float32_min = np.finfo(np.float32).min
    expected_output = np.clip(expected_output, float32_min, float32_max)

    output = rrcf._preprocess(data)
    np.testing.assert_array_equal(output, expected_output)


@given(
    st.dictionaries(
        st.text(),
        st.one_of(st.floats(width=32), st.integers()),
        min_size=1,
    )
)
def test_preprocess_missing_keys(data: dict[str, int | float]):
    rrcf = RobustRandomCutForest()
    rrcf._preprocess(data)

    input_data = {}
    output = rrcf._preprocess(input_data)
    assert np.count_nonzero(output) == 0


@given(
    st.dictionaries(
        st.text(),
        st.one_of(st.floats(width=32), st.integers()),
        min_size=1,
    )
)
def test_preprocess_unexpected_keys(data: dict[str, int | float]):
    rrcf = RobustRandomCutForest()
    rrcf._preprocess(data)

    new_key = "".join([*data, "abc"])
    data.update({new_key: 0.0})

    with pytest.raises(ValueError, match="Unseen features:"):
        rrcf._preprocess(data)


def test_preprocess_initial_keys():
    rrcf = RobustRandomCutForest()
    input_data = {"a": 1.0, "b": 2.0, "c": 3.0}
    rrcf._preprocess(input_data)
    assert rrcf._keys == ["a", "b", "c"]


def test_preprocess_nan_handling():
    rrcf = RobustRandomCutForest()
    input_data = {"a": 1.0, "b": np.nan, "c": 3.0}
    expected_output = np.array([1.0, 0.0, 3.0], dtype=np.float64)
    output = rrcf._preprocess(input_data)
    np.testing.assert_array_equal(output, expected_output)


@given(
    st.dictionaries(
        st.text(),
        st.one_of(st.floats(width=32), st.integers()),
        min_size=1,
    )
)
def test_learn_one_valid_input(data: dict[str, int | float]):
    rrcf = RobustRandomCutForest()
    initial_index = rrcf._index
    rrcf.learn_one(data)
    assert rrcf._index == initial_index + 1
    assert rrcf._shingle is not None
    assert len(rrcf._shingle) == rrcf.shingle_size
    for tree in rrcf.forest:
        assert len(tree.leaves) <= rrcf.tree_size


def test_learn_one_empty_input():
    rrcf = RobustRandomCutForest()
    initial_index = rrcf._index
    rrcf.learn_one({})
    assert rrcf._index == initial_index
    assert rrcf._shingle is None


def test_learn_one_initial_shingle():
    rrcf = RobustRandomCutForest()
    input_data = {"a": 1.0, "b": 2.0, "c": 3.0}
    rrcf.learn_one(input_data)
    assert rrcf._shingle is not None
    assert len(rrcf._shingle) == rrcf.shingle_size


def test_learn_one_tree_insertion():
    rrcf = RobustRandomCutForest()
    input_data = {"a": 1.0, "b": 2.0, "c": 3.0}
    rrcf.learn_one(input_data)
    for tree in rrcf.forest:
        assert len(tree.leaves) > 0


@given(
    st.dictionaries(
        st.one_of(st.floats(), st.integers(), st.text(), st.binary()),
        st.one_of(st.floats(width=32), st.integers()),
        min_size=1,
    )
)
def test_score_one_valid_input(data: dict[str, int | float]):
    rrcf = RobustRandomCutForest()
    rrcf.learn_one(data)
    score = rrcf.score_one(data)
    assert isinstance(score, float)
    assert score >= 0.0


def test_score_one_empty_input():
    rrcf = RobustRandomCutForest()
    score = rrcf.score_one({})
    assert score == 0.0


def test_score_one_tree_insertion():
    rrcf = RobustRandomCutForest()
    input_data = {"a": 1.0, "b": 2.0, "c": 3.0}
    rrcf.learn_one(input_data)
    initial_leaves = [len(tree.leaves) for tree in rrcf.forest]
    score = rrcf.score_one(input_data)

    assert isinstance(score, float)
    for initial, tree in zip(initial_leaves, rrcf.forest, strict=True):
        assert len(tree.leaves) == initial


@given(
    st.lists(
        st.fixed_dictionaries(
            {
                "a": st.one_of(st.floats(), st.integers()),
                "b": st.one_of(st.floats(), st.integers()),
                "c": st.one_of(st.floats(), st.integers()),
            },
        ),
        min_size=1,
        max_size=10,
    )
)
@settings(deadline=500)
def test_learn_score_multiple(data: list[dict[str, int | float]]):
    rrcf = RobustRandomCutForest()

    try:
        for d in data:
            _ = rrcf.score_one(d)
            rrcf.learn_one(d)
    except Exception as e:  # pragma: no cover
        pytest.fail(f"Exception raised: {e}")


@given(
    data=st.lists(
        st.fixed_dictionaries(
            {
                "a": st.one_of(st.floats(), st.integers()),
                "b": st.one_of(st.floats(), st.integers()),
                "c": st.one_of(st.floats(), st.integers()),
            },
        ),
        min_size=20,
        max_size=20,
    ),
    shingle_size=st.integers(min_value=1, max_value=4),
)
@settings(deadline=None)
def test_many_data(data: list[dict], shingle_size: int):
    rrcf = RobustRandomCutForest(num_trees=2, tree_size=4, shingle_size=shingle_size)

    try:
        for d in data:
            _ = rrcf.score_one(d)
            rrcf.learn_one(d)
    except Exception as e:  # pragma: no cover
        pytest.fail(f"Exception raised: {e}")


@given(
    num_trees=st.integers(min_value=1, max_value=100),
    tree_size=st.integers(min_value=1, max_value=256),
    shingle_size=st.integers(min_value=1, max_value=8),
    data=st.lists(
        st.dictionaries(
            st.one_of(st.floats(), st.integers(), st.text(), st.binary()),
            st.one_of(st.floats(), st.integers()),
        ),
        min_size=1,
        max_size=5,
    ),
)
@settings(deadline=None)
def test_rrcf_init_score_learn(
    num_trees: int,
    tree_size: int,
    shingle_size: int,
    data: list[dict],
):
    assume(len(data) > 0 and len(data[0]) > 0)

    rrcf = RobustRandomCutForest(
        num_trees=num_trees,
        tree_size=tree_size,
        shingle_size=shingle_size,
    )

    assert len(rrcf.forest) == num_trees
    assert rrcf.tree_size == tree_size
    assert rrcf.shingle_size == shingle_size
    assert rrcf._index == 0
    assert rrcf._keys is None
    assert rrcf._shingle is None

    all_keys = {key for d in data for key in d}
    sample = {k: 0.0 for k in all_keys}
    rrcf._preprocess(sample)

    assert rrcf._keys is not None

    scores = []
    for d in data:
        score = rrcf.score_one(d)
        scores.append(score)
        rrcf.learn_one(d)

    assert rrcf._index == len(data)
    assert rrcf._shingle is not None
    assert len(rrcf._shingle) == shingle_size

    for tree in rrcf.forest:
        assert len(tree.leaves) <= tree_size

    for score in scores:
        assert isinstance(score, float)
        assert score >= 0.0



from __future__ import annotations

import pickle
from uuid import uuid4
from tempfile import TemporaryDirectory
from pathlib import Path

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from river_rrcf.rrcf import RobustRandomCutForest


@pytest.mark.parametrize(
    "protocol",
    [*range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL + 1)],
)
@given(
    num_trees=st.integers(min_value=1, max_value=8),
    tree_size=st.integers(min_value=1, max_value=32),
    shingle_size=st.integers(min_value=1, max_value=4),
    data=st.lists(
        st.fixed_dictionaries(
            {
                "a": st.one_of(st.floats(), st.integers()),
                "b": st.one_of(st.floats(), st.integers()),
                "c": st.one_of(st.floats(), st.integers()),
            },
        ),
        min_size=1,
        max_size=6,
    ),
    random_state=st.one_of(st.none(), st.integers(min_value=0, max_value=1234567891)),
)
@settings(deadline=None)
def test_pickle(  # noqa: PLR0913 too-many-arguments
    num_trees: int,
    tree_size: int,
    shingle_size: int,
    data: list[dict],
    random_state: int | None,
    protocol: int,
):
    rrcf = RobustRandomCutForest(
        num_trees=num_trees,
        tree_size=tree_size,
        shingle_size=shingle_size,
        random_state=random_state,
    )

    for d in data[:-1]:
        rrcf.learn_one(d)

    with TemporaryDirectory() as tmp_dir:
        tmp_file = Path(tmp_dir, f"{uuid4()}.pkl")

        with tmp_file.open("wb") as f:
            pickle.dump(rrcf, f, protocol=protocol)

        with tmp_file.open("rb") as f:
            rrcf_loaded: RobustRandomCutForest
            rrcf_loaded = pickle.load(f)  # noqa: S301 suspicious-pickle-usage

        assert isinstance(rrcf_loaded, RobustRandomCutForest)
        dd = data[-1]
        assert rrcf_loaded.score_one(dd) == rrcf.score_one(dd)


@given(
    num_trees=st.integers(min_value=1, max_value=8),
    tree_size=st.integers(min_value=1, max_value=32),
    shingle_size=st.integers(min_value=1, max_value=4),
    data=st.lists(
        st.fixed_dictionaries(
            {
                "a": st.one_of(st.floats(), st.integers()),
                "b": st.one_of(st.floats(), st.integers()),
                "c": st.one_of(st.floats(), st.integers()),
            },
        ),
        min_size=1,
        max_size=6,
    ),
    random_state=st.one_of(st.none(), st.integers(min_value=0, max_value=1234567891)),
)
@settings(deadline=None)
def test_clone(
    num_trees: int,
    tree_size: int,
    shingle_size: int,
    data: list[dict],
    random_state: int | None,
):
    rrcf = RobustRandomCutForest(
        num_trees=num_trees,
        tree_size=tree_size,
        shingle_size=shingle_size,
        random_state=random_state,
    )

    for d in data[:-1]:
        rrcf.learn_one(d)

    rrcf_clone = rrcf.clone(include_attributes=True)
    assert isinstance(rrcf_clone, RobustRandomCutForest)
    dd = data[-1]
    assert rrcf_clone.score_one(dd) == rrcf.score_one(dd)
